/*
 * file PropertyRequestItem.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM
 *
 * (c) Copyright IBM Corporation 2004, 2008.  All Rights Reserved. 
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or  
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp. 
 */
package javax.wvcm;

import java.text.MessageFormat;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import javax.wvcm.PropertyNameList.PropertyName;


/**
 * A specification of the types of objects can be used to create a PropertyRequest.
 * 
 * @since 1.0
 */
public interface PropertyRequestItem { 

    /**
     * A map of PropertyName to PropertyRequest objects, used to specify one or more
     * properties whose values are to be read from a resource.
     * <p>
     * If the property value is itself a resource (or a data structure containing resources),
     * then the PropertyRequest value of the entry for the PropertyName of that property
     * specifies the properties that should be retrieved for that resource.
     * Such a request can be constructed using the {@link PropertyName#nest} method.
     * <p>
     * For example, the following code fragment creates a PropertyRequest that
     * specifies a request for the {@link Resource#CREATOR_DISPLAY_NAME},
     * {@link ControllableResource#CHECKED_IN}, and
     * {@link Resource#LAST_MODIFIED} properties, as well as the
     * {@link Version#VERSION_NAME} and {@link Resource#CREATION_DATE}
     * properties of the resource that is the value of the
     * {@link ControllableResource#CHECKED_IN} property:
     * 
     * <pre>
     *  PropertyRequest props = new PropertyRequest(
     *      Resource.CREATOR_DISPLAY_NAME,
     *      ControllableResource.CHECKED_IN.nest(
     *           Version.VERSION_NAME,
     *           Resource.CREATION_DATE),
     *      Resource.LAST_MODIFIED);
     * </pre>
     */
    public class PropertyRequest implements PropertyRequestItem, Feedback
    {

        private Map<PropertyName<?>, PropertyRequest> _itemMap;

        /**
         * A cached array form of the property request.
         */
        private NestedPropertyName<?>[] _itemArray;

        private static final NestedPropertyName<?>[] EMPTY_PNA =
            new NestedPropertyName[0];

        /**
         * An empty PropertyRequest. 
         */
        public static final PropertyRequest EMPTY = 
            new PropertyRequest();

        private static final Map<PropertyName<?>, PropertyRequest> EMPTY_MAP =
            Collections.emptyMap();
        
        /**
         * Constructs a PropertyRequest from an array of PropertyRequestItem objects.
         * If there are two entries for the same PropertyName, the PropertyRequest values
         * for those entries are logically merged.
         * 
         * @param items The array of PropertyRequestItem objects that are to
         *   be combined into a new PropertyRequest. 
         *   A <b>null</b> or empty array produces an empty PropertyRequest.
         */
        public PropertyRequest(PropertyRequestItem... items)
        {
            this(getItemMap(items));
        }

        /**
         * Constructs a PropertyRequest from an array of PropertyRequestItem objects.
         * If there are two entries for the same PropertyName, the PropertyRequest values
         * for those entries are logically merged.
         * 
         * @param items The array of PropertyRequestItem objects that are to
         *   be combined into a new PropertyRequest. 
         *   A <b>null</b> or empty array produces an empty PropertyRequest.
         */
        public PropertyRequest(PropertyName<?>[] items)
        {
            this((PropertyRequestItem[]) items);
        }

        /**
         * Constructs a PropertyRequest from an array of PropertyRequestItem objects.
         * If there are two entries for the same PropertyName, the PropertyRequest values
         * for those entries are logically merged.
         * 
         * @param items The array of PropertyRequestItem objects that are to
         *   be combined into a new PropertyRequest. 
         *   A <b>null</b> or empty array produces an empty PropertyRequest.
         */
        public PropertyRequest(NestedPropertyName<?>[] items)
        {
            this((PropertyRequestItem[]) items);
        }

        /**
         * Constructs a PropertyRequest from a <PropertyName, PropertyRequest> map.
         */
        private PropertyRequest(Map<PropertyName<?>, PropertyRequest> itemMap)
        {
            _itemMap = itemMap == null || itemMap.isEmpty()? null
            :Collections.unmodifiableMap(itemMap);
        }
        
        private static Map<PropertyName<?>, PropertyRequest>
        getItemMap(PropertyRequestItem... items)
        {
            if (items == null || items.length == 0) {
                return null;
            } else {
                Map<PropertyName<?>, PropertyRequest> itemMap 
                = new HashMap<PropertyName<?>, PropertyRequest>(items.length);
                
                for (PropertyRequestItem request: items) {
                    if (request == null) {
                        // ignore
                    } else if (request instanceof PropertyName) {
                        put(itemMap, (PropertyName<?>)request, null);
                    } else if (request instanceof NestedPropertyName) {
                        NestedPropertyName<?> npn = (NestedPropertyName<?>)request;
                        put(itemMap, npn.getRoot(), npn.getNested());
                    } else if (request instanceof PropertyNameList) {
                        PropertyNameList list = (PropertyNameList) request;

                        for (PropertyName<?> name : list.getPropertyNames())
                            put(itemMap, name, null);
                    } else if (request instanceof PropertyRequest) {
                        Map<PropertyName<?>, PropertyRequest> map =
                            ((PropertyRequest) request).toMap();

                        for (Entry<PropertyName<?>, PropertyRequest> entry : map.entrySet())
                            put(itemMap, entry.getKey(), entry.getValue());
                    } else {
                        throw new UnsupportedOperationException(request.getClass().getName() 
                                + " cannot be a PropertyRequest item"); //$NON-NLS-1$
                    }
                }
                
                return itemMap.isEmpty()? null: itemMap;
            }
        }

        private static void put(Map<PropertyName<?>, PropertyRequest> itemMap, 
                         PropertyName<?> propertyName, 
                         PropertyRequest propertyRequest) {
            PropertyRequest pr = itemMap.get(propertyName);
            itemMap.put(propertyName, merge(pr, propertyRequest));
        }

        private static PropertyRequest merge(PropertyRequest pr1, PropertyRequest pr2) {
            if (pr1 == null) {
                return pr2;
            }
            if (pr2 == null) {
                return pr1;
            }
            Map<PropertyName<?>, PropertyRequest> itemMap 
                = new HashMap<PropertyName<?>, PropertyRequest>(pr1.size() + pr2.size());
            Iterator<PropertyName<?>> keys = pr1.toMap().keySet().iterator();
            while (keys.hasNext()) {
                PropertyName<?> pName = keys.next();
                put(itemMap, pName, pr1.toMap().get(pName));
            }
            keys = pr2.toMap().keySet().iterator();
            while (keys.hasNext()) {
                PropertyName<?> pName = keys.next();
                put(itemMap, pName, pr2.toMap().get(pName));
            }
            PropertyRequest result = new PropertyRequest();
            result._itemMap = Collections.unmodifiableMap(itemMap);
            return result;
        }

        public Map<PropertyName<?>, PropertyRequest> toMap()
        {
            return (_itemMap == null) ? EMPTY_MAP : _itemMap;
        }

        /**
         * Returns an array containing all of the elements in this PropertyRequest.
         */
        public NestedPropertyName<?>[] toArray()
        { 
            if (_itemMap == null) {
                return EMPTY_PNA;
            }
            if (_itemArray == null) {
                NestedPropertyName<?>[] itemArray = new NestedPropertyName<?>[_itemMap.size()];
                int i = 0;
                for (PropertyName<?> propName: _itemMap.keySet()) {
                    PropertyRequest propRequest = _itemMap.get(propName);
                    itemArray[i++] = propName.nest(propRequest);
                }
                
                _itemArray = itemArray;
            }
            return _itemArray;
        }

        /**
         * @see java.util.Map#isEmpty
         */
        public boolean isEmpty()
        {
            return toMap().isEmpty();
        }

        /**
         * @see java.util.Map#size
         */
        public int size()
        {
            return toMap().size();        
        }

        /**
         * @see java.util.Map#get
         */
        public PropertyRequest get(PropertyName<?> key)
        {
            return toMap().get(key);        
        }

        @Override
        public boolean equals(Object object)
        {
            if (object == null || getClass() != object.getClass()) {
                return false;
            }
            return toMap().equals(((PropertyRequest)object).toMap());
        }

        @Override
        public int hashCode()
        {
            return toMap().hashCode();
        }

        @Override
        public String toString()
        {
            StringBuffer builder = new StringBuffer();
            NestedPropertyName<?>[] items = toArray();
            builder.append('[');
            for (int i=0; i<items.length; i++) {
                if (builder.length() > 1)
                    builder.append(", "); //$NON-NLS-1$
                builder.append(items[i].toString());
            }
            builder.append(']');

            return builder.toString();
        }

        public PropertyRequest getPropertyRequestForModified() {
            return null;
        }

        public PropertyRequest getPropertyRequestForResult() {
            return this;
        }

        public boolean isAbortRequested() {
            return false;
        }

        public Feedback nest() {
            return EMPTY;
        }

        public Feedback nest(PropertyRequest propertyRequest) {
            return propertyRequest;
        }

        public Feedback nest(int percentCompleted) {
            return EMPTY;
        }

        public Feedback nest(PropertyRequest resultPropertyRequest, int percentCompleted) {
            return resultPropertyRequest;
        }

        public String format(String fmt, Object...arguments) {
            return MessageFormat.format(fmt, arguments);
        }

        public void notifyActive(String message) {
        }

        public void notifyWarning(String message) {
        }

        public void notifyIsModified(Resource resource) {
        }

        public void notifyPercentComplete(int percentComplete) {
        }

    }

    /**
     * A NestedPropertyName consists of the name of a root property whose value is
     * desired and an optional request for nested properties of the resource (or
     * resources) referenced by the value of that root property.
     */
    public static class NestedPropertyName<T> implements PropertyRequestItem
    {
        final private PropertyNameList.PropertyName<T> _root;
        final private PropertyRequest _nested;
        private volatile int _hashCode = 0;

        /**
         * Constructs a NestedPropertyName from its constituent parts.
         * 
         * @param root The root PropertyName of the request; must not be <b>null</b>.
         * @param items The PropertyRequest for the value(s) of the root
         *            property; may be <b>null</b> or empty to request no
         *            nested properties.
         */
        public NestedPropertyName(PropertyName<T> root, 
                                  PropertyRequestItem... items) {

            if (root == null) {
                throw new UnsupportedOperationException
                ("The root PropertyName of a NestedPropertyName cannot be null."); //$NON-NLS-1$
            }

            _root = root;

            Map<PropertyName<?>, PropertyRequest> itemMap =
                PropertyRequest.getItemMap(items);

            _nested =
                itemMap == null || itemMap.isEmpty()
                    ? null
                    : new PropertyRequest(itemMap);
        }

        /**
         * Constructs a NestedPropertyName from its constituent parts.
         * 
         * @param root The root PropertyName of the request; must not be <b>null</b>.
         * @param items The PropertyRequest for the value(s) of the root
         *            property; may be <b>null</b> or empty to request no
         *            nested properties.
         */
        public NestedPropertyName(PropertyName<T> root, 
                                  PropertyName<?>[] items) 
        {
            this(root, (PropertyRequestItem[])items);
        }
        
        /**
         * Constructs a NestedPropertyName from its constituent parts.
         * 
         * @param root The root PropertyName of the request; must not be <b>null</b>.
         * @param items The PropertyRequest for the value(s) of the root
         *            property; may be <b>null</b> or empty to request no
         *            nested properties.
         */
        public NestedPropertyName(PropertyName<T> root, 
                                  NestedPropertyName<?>[] items) 
        {
            this(root, (PropertyRequestItem[])items);
        }
        
        /**
         * @return The root PropertyName of this nested property name object;
         * will never be <b>null</b>.
         */
        public PropertyName<T> getRoot() {
            return _root;
        }

        /**
         * @return The nested PropertyRequest, specifying the properties to be
         * requested from the resource or resources that are the value of the
         * root property name of this nested property name.
         */
        public PropertyRequest getNested() {
            return _nested==null ? PropertyRequest.EMPTY: _nested;
        }

        /**
         * Indicates whether some other object is "equal to" this one.
         * 
         * @param o the object to compare with.
         * @return true if and only if the specified object is a
         *         NestedPropertyName whose root property name and nested
         *         property name list values are equal to those of this object.
         */
        @Override
        public boolean equals(Object o)
        {
            if (o == null || this.getClass() != o.getClass()) {
                return false;
            }
            NestedPropertyName<?> npn = (NestedPropertyName<?>) o;
            return _root.equals(npn.getRoot())
            && getNested().equals(npn.getNested());
        }

        /**
         * Calculate a hash code value for the object.
         * 
         * @return the hash code for the object.
         */
        @Override
        public int hashCode() {
            if (_hashCode == 0) {
                int result = 17;
                result = 37 * result + _root.hashCode();
                result = 37 * result + getNested().hashCode();
                _hashCode = result;
            }
            return _hashCode;
        }

        /**
         * Returns a string representation of this NestedPropertyName suitable for
         * diagnostics.
         * 
         * @return The string representation of this NestedPropertyName formatted as
         * "property_name [first_nested_property_request, ... ,
         *         last_nested_property_request]"
         */
        @Override
        public String toString() {
            if (getNested().isEmpty())
                return _root.toString();
            else
                return _root.toString() + getNested().toString();
        }
    }	
}